#
# See the file "LICENSE" for the full license governing this code.
#
# encoding: UTF-8
require 'puppet_x/enterprisemodules/oracle/schemas'
require 'puppet_x/enterprisemodules/extend_hash'
# rubocop: disable Metrics/MethodLength

newproperty(:logfile_groups, :array_matching => :all) do
  include EasyType
  include EasyType::ArrayProperty
  include Puppet_X::EnterpriseModules::Oracle::Access
  include Puppet_X::EnterpriseModules::Oracle::Schemas
  include Puppet_X::EnterpriseModules::ExtendHash
  include Puppet_X::EnterpriseModules::Oracle::Information

  # Docs
  desc <<-EOD
    Specify the logfile groups.
    Array of hashes:
      [
        {file_name => '<abs_path_to_file|ASM diskgroup>',  # String, optional when using OMF
         group     => 1,                                   # Integer, optional. When not specified then this property behaves like a parameter(i.e. will not be managed).
         size      => '1G'                                 # String, optional default 100M
         thread    => 1,                                   # Integer, optional default 1
         type      => '<ONLINE|STANDBY>'                   # String, optional default 'ONLINE'
         reuse     => <true|false>                         # Boolean, optional
        }
      ]

    Use this property to specify all attributes of the logfile groups.
    The most simple form is to specify only the 'group' attribute. This
    requires that OMF is used. The size will be 100M by default and the
    logfile member(s) will be placed in the location that is specified by
    the database parameters db_create_online_log_dest_[1..5] or
    db_create_file_dest and/or db_recovery_file_dest:

        ora_database{'dbname':
          ...
          logfile_groups => [
              {group => 1, reuse => true},
              {group => 2, reuse => true},
              {group => 3, reuse => true},
            ],
        }

    When you want to specify more logfiles per loggroup, use specific log group
    numbers, assign them to a specific thread and/or create standby logfiles,
    you need to use the extended implementation:

        ora_database{'dbname':
          ...
          logfile_groups => [
              {group => 10, file_name => '/path/to/redo10a.log', size => '10M', reuse => true},
              {group => 10, file_name => '/path/to/redo10b.log', size => '10M', reuse => true},
              {group => 20, file_name => '/path/to/redo20a.log', size => '10M', reuse => true},
              {group => 20, file_name => '/path/to/redo20b.log', size => '10M', reuse => true},
              {group => 30, file_name => '/path/to/redo30a.log', size => '10M', thread => 2, reuse => true},
              {group => 30, file_name => '/path/to/redo30b.log', size => '10M', thread => 2, reuse => true},
              {group => 40, file_name => '/path/to/redo40a.log', size => '10M', thread => 2, reuse => true},
              {group => 40, file_name => '/path/to/redo40b.log', size => '10M', thread => 2, reuse => true},
              {group => 50, file_name => '/path/to/redo_stby50a.log', size => '10M', type => 'STANDBY', reuse => true},
              {group => 50, file_name => '/path/to/redo_stby50b.log', size => '10M', type => 'STANDBY', reuse => true},
              {group => 60, file_name => '/path/to/redo_stby60a.log', size => '10M', type => 'STANDBY', reuse => true},
              {group => 60, file_name => '/path/to/redo_stby60b.log', size => '10M', type => 'STANDBY', reuse => true},
              {group => 70, file_name => '/path/to/redo_stby70a.log', size => '10M', type => 'STANDBY', reuse => true},
              {group => 70, file_name => '/path/to/redo_stby70b.log', size => '10M', type => 'STANDBY', reuse => true},
              {group => 80, file_name => '/path/to/redo_stby80a.log', size => '10M', thread => 2, type => 'STANDBY', reuse => true},
              {group => 80, file_name => '/path/to/redo_stby80b.log', size => '10M', thread => 2, type => 'STANDBY', reuse => true},
              {group => 90, file_name => '/path/to/redo_stby90a.log', size => '10M', thread => 2, type => 'STANDBY', reuse => true},
              {group => 90, file_name => '/path/to/redo_stby90b.log', size => '10M', thread => 2, type => 'STANDBY', reuse => true},
              {group => 100, file_name => '/path/to/redo_stby100a.log', size => '10M', thread => 2, type => 'STANDBY', reuse => true},
              {group => 100, file_name => '/path/to/redo_stby100b.log', size => '10M', thread => 2, type => 'STANDBY', reuse => true},
            ],
        }

  EOD

  data_type("Optional[Array[Hash[String[1], Any]]]")

  munge do |orig_value|
    orig_value['size'] = munge_size(orig_value['size'])
    orig_value
  end

  #
  # This is a bit of hack. We would like to use the easy_type
  # size munger out of it's original context. But since the
  # Puppet 6 data types, the unsafe_munge falls back to a super class
  # for some cases. To make sure the super class is there,
  # we define it up front
  #
  class T
    def unsafe_munge(value)
      value
    end
  end

  def munge_size(size)
    object = T.new
    size ||= '100m'
    object.extend(EasyType::Mungers::Size)
    object.unsafe_munge(size)
  end

  to_translate_to_resource do |resource, raw_resource|
    return [] if resource['STATE'] != 'running'
    sid = resource['SID']
    command = <<-EOF
      select lf.group# as \"group\"
           , lf.member as \"file_name\"
           , l.thread# as \"thread\"
           , l.bytes as \"size\"
           , lf.type as \"type\"
        from v$logfile lf
           , v$log l
       where lf.group# = l.group#
       union
      select lf.group# as \"group\"
           , lf.member as \"file_name\"
           , l.thread# as \"thread\"
           , l.bytes as \"size\"
           , lf.type as \"type\"
        from v$logfile lf
           , v$standby_log l
       where lf.group# = l.group#
                 EOF
    values = sql(command, :sid => sid)
    all_values = values.each { |v| v.delete('SID') }
    all_values.map { |h| Hash[h.map { |k, v| (k == 'group' || k == 'thread' || k == 'size') ? [k, v.to_i] : [k, v] }.sort] }
  end

  VALIDATE_LOGFILE_GROUP = Puppet_X::EnterpriseModules::Oracle::Schemas::LOGFILEGROUP

  def validate(value)
    value = [value] if value.is_a?(Hash) # ensure, it is an array
    value.each { |v| ClassyHash.validate_strict(v, VALIDATE_LOGFILE_GROUP) }
  end

  def insync?(is)
    if is.to_s == 'absent'
      should.empty?
    elsif should.to_s == 'absent'
      is.empty?
    else
      numbered_groups = numbered_groups_value(should)
      non_numbered_groups = non_numbered_groups_value(should)
      fail 'cannot mix numbered and non-numbered groups' if numbered_groups.any? && non_numbered_groups.any?
      return true if non_numbered_groups.any?
      converted_is = convert(is)
      converted_should = convert(should)
      # TODO: Check if this(sort_by) should be done in easy_type too
      converted_is.sort_by { |h| h.keys[0] } == converted_should.sort_by { |h| h.keys[0] }
    end
  end

  def change_to_s(_currentvalue, _newvalue)
    converted_is = convert(provider.logfile_groups)
    converted_should = convert(should)
    add_lfg = subtract_once(converted_should, converted_is)
    rm_lfg = subtract_once(converted_is, converted_should)
    current_lf_size = provider.logfile_groups.map { |h| h['size'] }.uniq
    new_lf_size = should.map { |h| h['size'] }.uniq
    if new_lf_size.join.to_i != current_lf_size.join.to_i
      "resizing logfiles #{provider.logfile_groups} to size #{new_lf_size.first}"
    elsif rm_lfg.any? && add_lfg.any?
      "removing #{rm_lfg} adding #{add_lfg}"
    elsif rm_lfg.any?
      "removing #{rm_lfg}"
    elsif add_lfg.any?
      "adding #{add_lfg}"
    end
  end

  on_modify do
    modify
  end

  after_create do
    is = should.select { |x| x['type'] == 'ONLINE' && x['thread'] == 1 }
    modify(is, should)
  end

  def modify(currentvalue = provider.logfile_groups, newvalue = should)
    converted_is = convert(currentvalue)
    converted_should = convert(newvalue)
    add_lfg = subtract_once(converted_should, converted_is)
    rm_lfg = subtract_once(converted_is, converted_should)

    # Check current and specified logfile sizes
    current_lf_size = currentvalue.map { |h| h['size'] }.uniq
    new_lf_size = newvalue.map { |h| h['size'] }.uniq
    fail "all logfiles should have the same size current_sizes = (#{current_lf_size.join(', ')}), new_sizes = (#{new_lf_size.join(', ')})" if new_lf_size.length > 1
    fail 'your config would cause all logfiles to be dropped, cannot continue' if rm_lfg.length == converted_is.length && new_lf_size.join.to_i == current_lf_size.join.to_i

    # Either resize the redologs or remove or add logfiles in current run
    if new_lf_size.join.to_i != current_lf_size.join.to_i
      new_size = new_lf_size.join.to_i
      command = template('ora_config/ora_database/resize_redologs.sql.erb', binding)
      sql(command, :sid => instance)
    else
      if rm_lfg.any?
        # Group by 'group'
        group_rm_lfg = rm_lfg.group_by { |element| element['group'] }
        # Loop over groups
        group_rm_lfg.each do |group, hash|
          delete_logs_for_group = hash.count { |h| h['group'] == group }
          current_logs_for_group = converted_is.count { |h| h['group'] == group }
          # Check if we have to drop member(s) or group
          command = if delete_logs_for_group < current_logs_for_group
                      # At least one member remaining for group
                      real_logs = hash.map { |h| currentvalue.select { |x| x['group'] == group && x['thread'] == h['thread'] && x['type'] == h['type'] && x['file_name'].include?(h['file_name']) } }.flatten.uniq
                      selected_logs = real_logs[0..delete_logs_for_group - 1]
                      "alter database drop logfile member #{logfile_member_clause(selected_logs)}"
                    else
                      # Drop the group
                      "alter database drop logfile group #{group}"
                    end
          begin
            sql(command, :sid => instance)
          rescue
            # Switch logfile and checkpoint when logfile group (member) cannot be dropped
            switch_command = <<-EOF
              alter system switch logfile;
              alter system archive log group #{group};
              alter system checkpoint global;
              #{command}
            EOF
            sql(switch_command, :sid => instance)
          end
        end
      end
      if add_lfg.any?
        # Check the different logfile types we have to add
        add_lfg_type = add_lfg.map { |h| h['type'] }.uniq
        # Loop over types
        add_lfg_type.each do |type|
          type_spec = 'standby' if type == 'STANDBY'
          # Select the current type
          add_lfg_for_type = add_lfg.select { |hash| hash['type'] == type }
          # Next if there aren't any
          next unless add_lfg_for_type.any?
          # Check the different threads we have to add for
          threads = add_lfg_for_type.map { |h| h['thread'] }.uniq
          # Loop over threads, because one command cannot add logfiles for multiple threads
          threads.each do |thread|
            # Check which logs we have to add for thread
            add_lfg_type_for_thread = add_lfg_for_type.select { |hash| hash['thread'] == thread }
            # Group by 'group'
            group_add_lfg_type_for_thread = add_lfg_type_for_thread.group_by { |element| element['group'] }
            # Loop over groups
            group_add_lfg_type_for_thread.each do |group, hash|
              if converted_is.any? { |h| h['group'] == group }
                # There is already a group so we have to add member(s)
                fail 'unable to add logfile members when OMF is used on filesystem' if hash.first['file_name'].nil? && db_cf_dest =~ %r{^/(.*)$}
                sql("alter database add #{type_spec} logfile member #{logfile_member_clause(hash)} reuse to group #{group}", :sid => instance)
              else
                # Add a new online logfile group
                sql("alter database add #{type_spec} #{logfile_clause(hash)}", :sid => instance)
              end
            end
          end
        end
      end
    end
  end

  def value
    # only the online logs for thread 1 are needed here
    should_with_defaults = add_defaults(should)
    online_logs = should_with_defaults.select { |h| h['type'] != 'STANDBY' && h['thread'] == 1 }
    logfile_clause(online_logs, true)
  end

  def logfile_clause(logfile_array, create_db = false)
    numbered_groups = numbered_groups_value(logfile_array)
    non_numbered_groups = non_numbered_groups_value(logfile_array)
    fail 'cannot mix numbered and non-numbered groups' if numbered_groups.any? && non_numbered_groups.any?
    command_segment = []
    if non_numbered_groups.any?
      thread = non_numbered_groups.map { |h| h['thread'] }.uniq.first
      non_numbered_groups.each_index do |index|
        group = index + 1
        entry = non_numbered_groups[index]
        command_segment << "group #{group} #{file_specification(entry)}"
      end
      if create_db
        "logfile #{command_segment.join(', ')}"
      else
        "logfile thread #{thread} #{command_segment.join(', ')}"
      end
    elsif numbered_groups.any?
      group_numbers = numbered_groups.collect { |e| e['group'] }.uniq
      thread = numbered_groups.map { |h| h['thread'] }.uniq.first
      group_numbers.each do |group|
        files = numbered_groups.map { |e| e['file_name'] if e['group'] == group }.compact
        entry = numbered_groups.map do |e|
          e['reuse'] = true if normal_file?(e['file_name']) && !e['file_name'].nil?
          e if e['group'] == group
        end.compact.first
        file_segments = files.collect { |f| "'#{f}'" }
        command_segment << if files.empty?
                             "group #{group} #{file_characteristics(entry)} "
                           elsif files.count == 1
                             "group #{group} #{file_segments.join(', ')} #{file_characteristics(entry)} "
                           else
                             "group #{group} (#{file_segments.join(', ')}) #{file_characteristics(entry)} "
                           end
      end
      if create_db
        "logfile #{command_segment.join(', ')}"
      else
        "logfile thread #{thread} #{command_segment.join(', ')}"
      end
    else
      ''
    end
  end

  def numbered_groups_value(array_value)
    array_value.map { |e| e unless e['group'].nil? }.compact
  end

  def non_numbered_groups_value(array_value)
    array_value - numbered_groups_value(array_value)
  end

  def add_defaults(input_value)
    # Add defaults for thread and type
    if input_value.is_a?(Array)
      input_value.map do |h|
        h['thread'] = 1 unless h.key?('thread')
        h['type'] = 'ONLINE' unless h.key?('type')
        h
      end
    else
      input_value['thread'] = 1 unless input_value.key?('thread')
      input_value['type'] = 'ONLINE' unless input_value.key?('type')
      input_value
    end
  end

  def logfile_member_clause(hash_value)
    hash_value.collect { |h| "'#{h['file_name']}'" }.join(', ')
  end

  def convert(hash_value)
    if hash_value == provider.logfile_groups
      # dealing with 'is'
      regexp_asm = regexp_omf_asm
      regexp_fs = regexp_omf_fs
      hash_value.map do |hash|
        result = Hash[hash.map do |key, value|
          if key == 'file_name'
            if value =~ /^#{regexp_asm}.*group_\d+\.\d+\.\d+$/i
              # file_name is an Oracle Managed File in ASM, only leave +diskgroup in file_name
              [key, value.split('/')[0]]
            elsif value =~ /^#{regexp_fs.gsub('\\','\\\\\\')}.*o1_mf_\d+_.*\.log$/i
              # file_name is an Oracle Managed File on filesystem, add entry with nil so we can strip it later
              %w(nil nil)
            else
              [key, value]
            end
          else
            [key, value]
          end
        end]
        # Remove key 'nil'
        result.key?('nil') ? result.delete('nil') : result
        result
      end
    else
      # dealing with 'should'
      # Get the database parameters that could create logfiles without specifying 'file_name'
      dbcd1 = db_col_dest_1
      dbcd2 = db_col_dest_2
      dbcd3 = db_col_dest_3
      dbcd4 = db_col_dest_4
      dbcd5 = db_col_dest_5
      dbcd = db_cf_dest
      dbrd = db_rf_dest
      hashes = hash_value.map do |hash|
        hash['file_name'] = '' unless hash.key?('file_name')
        if !omf_file?(hash['file_name']) && hash['file_name'] != ''
          # Not an OMF, leave as is
          hash
        else
          # rubocop: disable Style/IfInsideElse
          # rubocop: disable Metrics/BlockNesting
          # return hash unless omf_file?(hash['file_name']) || hash['file_name'] == '' # Not an OMF, leave as is
          if dbcd1 != '' || dbcd2 != '' || dbcd3 != '' || dbcd4 != '' || dbcd5 != ''
            # One or multiple db_create_online_log_dest_? are set
            asm_col_dests = get_asm_dests([dbcd1, dbcd2, dbcd3, dbcd4, dbcd5])
            fs_col_dests = get_fs_dests([dbcd1, dbcd2, dbcd3, dbcd4, dbcd5])
            if asm_col_dests.any?
              # Add (multiple) file_name(s) to hash
              hash = add_file_name_to_hash(hash, asm_col_dests)
            elsif fs_col_dests.any?
              # Remove file_name if OMF on filesystem is used
              hash.delete('file_name')
              hash
            end
          elsif dbcd != '' || dbrd != ''
            # db_create_file_dest or db_recovery_file_dest is set
            # Check if they are in ASM or on filesystem
            asm_col_dests = get_asm_dests([dbcd, dbrd])
            fs_col_dests = get_fs_dests([dbcd, dbrd])
            if asm_col_dests.any?
              # Add (multiple) file_name(s) to hash
              hash = add_file_name_to_hash(hash, asm_col_dests)
            elsif fs_col_dests.any?
              # Remove file_name is OMF on filesystem is used
              hash.delete('file_name')
              hash
            end
          end
        end
        hash = add_defaults(hash)
        hash
      end
      if hash_value.length > hashes.uniq.flatten.length
        hashes.flatten
      else
        hashes.uniq.flatten
      end
    end
  end

  def subtract_once(array1, array2)
    counts = array2.inject(Hash.new(0)) { |a, e| a[e] += 1; a }
    array1.reject { |e| counts[e] -= 1 unless counts[e].zero? }
  end

  def get_asm_dests(array)
    # returns an array of destinations in ASM that are not nil
    array.select { |dest| dest.start_with?('+') }
  end

  def get_fs_dests(array)
    # returns an array of destinations on filesystem that are not nil
    array.select { |dest| !dest.to_s.empty? and !dest.start_with?('+') }
  end

  def omf_file?(file_name)
    # Check if file_name is an Oracle Managed File
    file = file_name.split('/')[-1]
    if file =~ /^group_\d+\.\d+\.\d+$/ || file =~ /^o1_mf_\d+_.*\.log$/
      true
    else
      false
    end
  end

  def normal_file?(file_name)
    # Check if file_name is a normal file
    file_name !~ /^\+.*\/group_\d+\.\d+\.\d+$/ || file !~ /^\/.*\/o1_mf_\d+_.*\.log$/ ? true : false
  end

  def add_file_name_to_hash(hash_value, destinations)
    destinations.map do |dest|
      hash = Hash[hash_value.map do |key, value|
                    [key, value]
                  end]
      hash['file_name'] = dest
      Hash[hash.sort]
    end
  end

  def db_cf_dest
    value_for_init_param('db_create_file_dest', instance)
  end

  def db_rf_dest
    value_for_init_param('db_recovery_file_dest', instance)
  end

  def db_col_dest_1
    value_for_init_param('db_create_online_log_dest_1', instance)
  end

  def db_col_dest_2
    value_for_init_param('db_create_online_log_dest_2', instance)
  end

  def db_col_dest_3
    value_for_init_param('db_create_online_log_dest_3', instance)
  end

  def db_col_dest_4
    value_for_init_param('db_create_online_log_dest_4', instance)
  end

  def db_col_dest_5
    value_for_init_param('db_create_online_log_dest_5', instance)
  end

  def regexp_omf_asm
    regexp_omf_asm = '('
    regexp_omf_asm << (db_cf_dest.include?('+') ? "#{db_cf_dest}|" : '')
    regexp_omf_asm << (db_rf_dest.include?('+') ? "#{db_rf_dest}|" : '')
    regexp_omf_asm << (db_col_dest_1.include?('+') ? "#{db_col_dest_1}|" : '')
    regexp_omf_asm << (db_col_dest_2.include?('+') ? "#{db_col_dest_2}|" : '')
    regexp_omf_asm << (db_col_dest_3.include?('+') ? "#{db_col_dest_3}|" : '')
    regexp_omf_asm << (db_col_dest_4.include?('+') ? "#{db_col_dest_4}|" : '')
    regexp_omf_asm << (db_col_dest_5.include?('+') ? "#{db_col_dest_5}|" : '')
    regexp_omf_asm = regexp_omf_asm[-1] == '|' ? regexp_omf_asm[0..-2] : regexp_omf_asm
    regexp_omf_asm << ')'
    regexp_omf_asm.gsub('+', '\\\\+')
  end

  def regexp_omf_fs
    if Puppet::Util::Platform.windows?
      file_dest_regex = %r{^[a-z]:(.*)$}i
    else
      file_dest_regex = %r{^/(.*)$}
    end
    regexp_omf_fs = '('
    regexp_omf_fs << (db_cf_dest =~ file_dest_regex ? "#{db_cf_dest}|" : '')
    regexp_omf_fs << (db_rf_dest =~ file_dest_regex ? "#{db_rf_dest}|" : '')
    regexp_omf_fs << (db_col_dest_1 =~ file_dest_regex ? "#{db_col_dest_1}|" : '')
    regexp_omf_fs << (db_col_dest_2 =~ file_dest_regex ? "#{db_col_dest_2}|" : '')
    regexp_omf_fs << (db_col_dest_3 =~ file_dest_regex ? "#{db_col_dest_3}|" : '')
    regexp_omf_fs << (db_col_dest_4 =~ file_dest_regex ? "#{db_col_dest_4}|" : '')
    regexp_omf_fs << (db_col_dest_5 =~ file_dest_regex ? "#{db_col_dest_5}|" : '')
    regexp_omf_fs = regexp_omf_fs[-1] == '|' ? regexp_omf_fs[0..-2] : regexp_omf_fs
    regexp_omf_fs << ')'
  end
end
