#
# See the file "LICENSE" for the full license governing this code.
#
newproperty(:size) do
  include EasyType
  include EasyType::Mungers::Size
  include TablespaceType
  include Puppet_X::EnterpriseModules::Oracle::Access
  include Puppet_X::EnterpriseModules::Oracle::DatafileProperty

  desc <<-EOD
    The size of the tablespace.
  EOD

  # Issue in Puppet causes external types to not always 
  # be resolved. Therefore we include the basic variant
  # data_type("Optional[Easy_type::Size]")
  data_type("Optional[Variant[Pattern[/^([0-9]*[k|K|m|M|g|G|t|T]|unlimited|UNLIMITED)$/], Integer]]")

  to_translate_to_resource do |raw_resource|
    raw_resource.column_data('BYTES').to_i
  end

  def insync?(is)
    if autoextend?
      Puppet.debug "#{path}: Puppet ignores the Tablespace[size] attribute when autoextend is on." if different?
      #
      # When autoextend is on, we expect Oracle to manage the size. So we won't do
      # anything with the size attribute when the resource already exists
      #
      true
    elsif resource.size < provider.size
      Puppet.warning "#{path}: Puppet ignores the Tablespace[size] attribute when the specified size is smaller than current size"
    else
      #
      # When there is no autoextend on, we just compare the values and pass it to oracle
      #
      is == should
    end
  end

  on_modify do
    provider.datafile.each do |df|
      sql("alter database #{file_type} '#{df}' resize #{value}", :sid => resource.sid)
    end
  end

  on_create do
    if resource[:datafile].nil?
      # num_datafiles specified
      if desired_num_datafiles == 0
        "#{tablespace_type} size #{resource[:size]}"
      elsif desired_num_datafiles == 1
        # with one datafile only add the size here. autoextend and next are added in their properties
        "#{tablespace_type}" + (1..resource[:num_datafiles].to_i).map { |_i| " size #{resource[:size]}" }.join(',')
      else
        # with multible datafiles the autoextend and next should be added here
        "#{tablespace_type}" + (1..resource[:num_datafiles].to_i).map { |_i| " size #{resource[:size]} #{autoextend_spec}" }.join(',')
      end
    else
      # datafile specified
      if df_array_size(resource.datafile) == 1
        # with one datafile only add the size here. autoextend and next are added in their properties
        "#{tablespace_type} '#{resource[:datafile].join}' size #{resource[:size]}"
      else
        # with multible datafiles the autoextend and next should be added here
        "#{tablespace_type}" + resource[:datafile].map {|df| " '#{df}' size #{resource[:size]} #{autoextend_spec}" }.join(',')
      end
    end
  end

  private

  def different?
    current_size != value
  end

  def equal?
    current_size == value
  end

  def current_size
    provider.get(:size)
  end

  def autoextend?
    provider.get(:autoextend).to_s == 'on'
  end

end
